/*

 Copyright 2013, SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.wrapper;

import com.sun.jna.Function;
import com.sun.jna.Memory;
import com.sun.jna.NativeLibrary;
import com.sun.jna.Pointer;
import com.sun.jna.ptr.PointerByReference;
import java.io.File;
import java.nio.ByteBuffer;
import java.nio.DoubleBuffer;
import java.nio.IntBuffer;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.UUID;
import org.javafmi.FmuLocation;
import org.javafmi.ModelDescription;
import org.javafmi.Status;
import static org.javafmi.Status.*;
import org.javafmi.modeldescription.modelvariables.ScalarVariable;
import org.javafmi.modeldescription.typedefinitions.EnumerationType;
import org.javafmi.operatingsystem.FmuFileExtractor;
import static org.javafmi.wrapper.FmuFunctionSuffix.*;
import org.javafmi.wrapper.datatypes.NativeBoolean;

public class FmuProxy implements FunctionalMockupInterface {

    private static final String MYME_TYPE = "application/x-fmu-sharedlibrary";
    private final String fmuLocation;
    private final String proxyID;
    private final String modelIdentifier;
    private final ModelDescription modelDescription;
    private final String libraryPath;
    private NativeLibrary libraryInstance;
    private boolean isInstanceVisible;
    private boolean isInstanceInteractive;
    private boolean isDoingLogging;
    private boolean isStopTimeDefined;
    private double startTime;
    private double stopTime;
    private double currentSimulationTime;
    private double timeout;
    private Pointer instantiatedSlave;
    private Map<String, Function> nativeFunctions;
    private IntBuffer valueReferenceBuffer;
    private DoubleBuffer outBufferForDouble;
    private IntBuffer outBufferForInt;
    private ByteBuffer outBufferForByte;
    private Size_T sizeT;
    private CallbackFunctionsStructure.ByValue callbacks;
    private final FmuFileExtractor fileExtractor;

    public FmuProxy(FmuLocation location) {
        this.callbacks = new CallbackFunctionsStructure.ByValue();
        this.fmuLocation = location.toString();
        this.fileExtractor = new FmuFileExtractor(new File(fmuLocation), String.valueOf(System.currentTimeMillis()));
        this.modelDescription = ModelDescriptionDeserializer.deserialize(fileExtractor.getModelDescriptionFile());
        modelDescription.buildAttributes();
        this.libraryPath = fileExtractor.getLibraryPath(modelDescription.getModelIdentifier());
        this.modelIdentifier = modelDescription.getModelIdentifier();
        this.proxyID = modelIdentifier + "_" + UUID.randomUUID().toString();
        this.isInstanceVisible = false;
        this.isInstanceInteractive = false;
        this.isDoingLogging = false;
        this.isStopTimeDefined = true;
    }

    @Override
    public void instantiateSlave(double timeout) {
        this.timeout = timeout;
        this.libraryInstance = NativeLibrary.getInstance(libraryPath);
        this.nativeFunctions = getNativeFunctions();
        this.instantiatedSlave = getPointerToLibrary();
        this.valueReferenceBuffer = IntBuffer.allocate(1);
        this.outBufferForInt = IntBuffer.allocate(1);
        this.outBufferForDouble = DoubleBuffer.allocate(1);
        this.outBufferForByte = ByteBuffer.allocate(1);
        this.sizeT = new Size_T(1);
    }

    private Pointer getPointerToLibrary() {
        Pointer pointerToLibrary = (Pointer) nativeFunctions.get(INSTANTIATE).invoke(Pointer.class, getParametersForInstantiate());
        if (pointerToLibrary == Pointer.NULL)
            throw new RuntimeException("Imposible to instantiate FMU" + modelIdentifier);
        return pointerToLibrary;
    }

    public Map<String, Function> getNativeFunctions() {
        LinkedHashMap<String, Function> functions = new LinkedHashMap<>();
        for (String suffix : FmuFunctionSuffix.getFunctionsSuffix())
            functions.put(suffix, getFunction(suffix));
        return functions;
    }

    @Override
    public Status initializeSlave(double startTime, double stopTime) {
        this.currentSimulationTime = startTime;
        this.startTime = startTime;
        this.stopTime = stopTime;
        return getStatus((Integer) nativeFunctions.get(INITIALIZE).invoke(Integer.class, getParametersForInitialize()));
    }

    @Override
    public Status terminateSlave() {
        return getStatus((Integer) nativeFunctions.get(TERMINATE).invoke(Integer.class, new Object[]{instantiatedSlave}));
    }

    @Override
    public Status resetSlave() {
        return getStatus((Integer) nativeFunctions.get(RESET).invoke(Integer.class, new Object[]{instantiatedSlave}));
    }

    @Override
    public void freeSlave() {
        nativeFunctions.get(FREE).invoke(new Object[]{instantiatedSlave});
        libraryInstance.dispose();
        fileExtractor.deleteTemporal();
    }

    @Override
    public Status doStep(double stepSize) {
        Status doStepStatus = getStatus(nativeFunctions.get(DO_STEP).invokeInt(getParametersForDoStep(stepSize)));
        if (doStepStatus.equals(Status.OK))
            currentSimulationTime += stepSize;
        return doStepStatus;
    }

    @Override
    public Status cancelStep() {
        return getStatus((Integer) nativeFunctions.get(CANCEL_STEP).invoke(Integer.class, new Object[]{instantiatedSlave}));
    }

    @Override
    public ScalarVariable getScalarVariable(String name) {
        return modelDescription.getScalarVariable(name);
    }

    @Override
    public Object getEnumeration(ScalarVariable modelVariable) {
        EnumerationType enumeration = (EnumerationType) modelVariable.getType();
        Integer enumerationIndex;
        EnumerationType enumarationType = (EnumerationType) modelDescription.getTypeFromTypeDefinition(enumeration.getDeclaredType());
        if (modelVariable.getVariability().equalsIgnoreCase("constant")) {
            enumerationIndex = enumeration.getStart();
            return enumarationType.getItems().get(enumerationIndex - 1).getName();
        }
        enumerationIndex = getInteger(modelVariable.getValueReference());
        return enumarationType.getItems().get(enumerationIndex - 1).getName();

    }

    @Override
    public Double getReal(Integer valueReference) {
        nativeFunctions.get(GET_REAL).invokeInt(getParametersForFmuGetterSetter(valueReference, outBufferForDouble));
        return outBufferForDouble.get(0);
    }

    @Override
    public int getInteger(Integer valueReference) {
        nativeFunctions.get(GET_INTEGER).invoke(getParametersForFmuGetterSetter(valueReference, outBufferForInt));
        return outBufferForInt.get(0);
    }

    @Override
    public boolean getBoolean(Integer valueReference) {
        nativeFunctions.get(GET_BOOLEAN).invoke(getParametersForFmuGetterSetter(valueReference, outBufferForByte));
        return NativeBoolean.getBooleanFor(outBufferForByte.get(0));
    }

    @Override
    public String getString(Integer valueReference) {
        PointerByReference outPointerByReference = new PointerByReference();
        nativeFunctions.get(GET_STRING).invoke(getParametersForFmuGetterSetter(valueReference, outPointerByReference));
        Pointer stringPointer = outPointerByReference.getValue();
        return (stringPointer != null) ? stringPointer.getString(0) : "";
    }

    @Override
    public Status setReal(Integer valueReference, Double newValue) {
        outBufferForDouble.put(0, newValue);
        return getStatus(nativeFunctions.get(SET_REAL).invokeInt(getParametersForFmuGetterSetter(valueReference, outBufferForDouble)));
    }

    @Override
    public Status setInteger(Integer valueReference, Integer newValue) {
        outBufferForInt.put(0, newValue);
        return getStatus((Integer) nativeFunctions.get(SET_INTEGER).invoke(Integer.class, getParametersForFmuGetterSetter(valueReference, outBufferForInt)));
    }

    @Override
    public Status setBoolean(Integer valueReference, Boolean newValue) {
        outBufferForByte.put(0, NativeBoolean.getValueFor(newValue));
        return getStatus(nativeFunctions.get(SET_BOOLEAN).invokeInt(getParametersForFmuGetterSetter(valueReference, outBufferForByte)));
    }

    @Override
    public Status setString(Integer valueReference, String newValue) {
        PointerByReference pointerByReference = new PointerByReference();
        Pointer reference = new Memory(newValue.length() + 1).share(0);
        reference.setString(0, newValue);
        pointerByReference.setValue(reference);
        return getStatus(nativeFunctions.get(SET_STRING).invokeInt(getParametersForFmuGetterSetter(valueReference, pointerByReference)));
    }

    @Override
    public ModelDescription getModelDescription() {
        return modelDescription;
    }

    @Override
    public boolean isInstanceVisible() {
        return isInstanceVisible;
    }

    @Override
    public void setIsInstanceVisible(boolean isInstanceVisible) {
        this.isInstanceVisible = isInstanceVisible;
    }

    @Override
    public boolean isInstanceInteractive() {
        return isInstanceInteractive;
    }

    @Override
    public void setIsInstanceInteractive(boolean isInstanceInteractive) {
        this.isInstanceInteractive = isInstanceInteractive;
    }

    @Override
    public boolean isDoingLogging() {
        return isDoingLogging;
    }

    @Override
    public Status setIsDoingLogging(boolean isDoingLogging) {
        this.isDoingLogging = isDoingLogging;
        return getStatus((Integer) nativeFunctions.get(SET_DEBUG).invoke(Integer.class, getParametersForSetDebugLogging()));
    }

    private Function getFunction(String suffix) {
        return libraryInstance.getFunction(modelIdentifier + suffix);
    }

    private Object[] getParametersForInstantiate() {
        return new Object[]{proxyID, modelDescription.getGuid(), fmuLocation, MYME_TYPE, timeout, NativeBoolean.getValueFor(isInstanceVisible), NativeBoolean.getValueFor(isInstanceInteractive), callbacks, NativeBoolean.getValueFor(isDoingLogging)};
    }

    private Object[] getParametersForInitialize() {
        return new Object[]{instantiatedSlave, startTime, NativeBoolean.getValueFor(isStopTimeDefined), stopTime};
    }

    private Object[] getParametersForDoStep(double stepSize) {
        return new Object[]{instantiatedSlave, currentSimulationTime, stepSize, (byte) 1};
    }

    public Object[] getParametersForFmuGetterSetter(Integer valueReference, Object outBuffer) {
        return new Object[]{instantiatedSlave, valueReferenceBuffer.put(0, valueReference), sizeT, outBuffer};
    }

    private Object[] getParametersForSetDebugLogging() {
        return new Object[]{instantiatedSlave, NativeBoolean.getValueFor(isDoingLogging)};
    }

    @Override
    public double getCurrentSimulationTime() {
        return currentSimulationTime;
    }

    public Function getNativeFunction(String function) {
        return nativeFunctions.get(function);
    }
}
